from __future__ import annotations

import re
from typing import List

from aasm.preprocessor.preprocessor_item import PreprocessorItem


class Macro(PreprocessorItem):
    def __init__(self, signature):
        super().__init__(signature)
        self.lines = []
        self.argument_regexes = []
        self.declare_line = -1
        self.name = ""
        self.expand_len = 0

    def __repr__(self):
        return f"Macro({self.name})"

    def expand(self, macro_args: List[str]) -> List[str]:
        expanded = []
        macro_args = [" " + arg + "," for arg in macro_args]
        substitutions = list(zip(macro_args, self.argument_regexes))
        for line in self.lines:
            expand_line = line
            for sub in substitutions:
                # expand_line = expand_line.replace(sub[1], sub[0])
                expand_line = re.sub(sub[1], sub[0], expand_line)
                expand_line = expand_line.rstrip(",")
            expanded.append(expand_line)
        return expanded

    def _create_regex(self, name: str) -> str:
        return f"[\\s,]{name}\\s*,?"

    def add_definition(self, definition: List[str], line_idx: int):
        # definition = [token for token in definition if token != '']
        self.name = definition[0]
        self.declare_line = line_idx
        self.argument_regexes = [self._create_regex(arg) for arg in definition[1:]]

    def add_line(self, line: str):
        self.lines.append(line)
        self.expand_len += 1



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, List


class MessageParam:
    def __init__(self, name: str):
        self.name: str = name
        self.value: str = ""
        self.is_value_set: bool = False

    def set_value(self, value: str) -> None:
        self.is_value_set = True
        self.value = value

    def print(self) -> None:
        print(f"MessageParam {self.name} = {self.value}")


class FloatParam(MessageParam):
    def __init__(self, name: str):
        super().__init__(name)

    def print(self) -> None:
        print(f"MessageFloatParam")
        super().print()


class ConnectionParam(MessageParam):
    def __init__(self, name: str):
        super().__init__(name)

    def print(self) -> None:
        print(f"MessageConnectionParam")
        super().print()


class Message:
    RESERVED_CONNECTION_PARAMS = ["sender"]
    RESERVED_TYPE_PARAMS = ["type", "performative"]

    def __init__(self, msg_type: str, msg_performative: str):
        self.type: str = msg_type
        self.performative: str = msg_performative
        self.float_params: Dict[str, FloatParam] = {}
        self.connection_params: Dict[str, MessageParam] = {}

    @property
    def param_names(self) -> List[str]:
        return [
            *Message.RESERVED_CONNECTION_PARAMS,
            *Message.RESERVED_TYPE_PARAMS,
            *list(self.float_params),
            *list(self.connection_params),
        ]

    @property
    def unset_params(self) -> List[str]:
        unset_params: List[str] = []

        for name, float_param in self.float_params.items():
            if not float_param.is_value_set:
                unset_params.append(name)

        for name, connection_param in self.connection_params.items():
            if not connection_param.is_value_set:
                unset_params.append(name)

        return unset_params

    def are_all_params_set(self) -> bool:
        return all(
            [float_param.is_value_set for float_param in self.float_params.values()]
            + [
                connection_param.is_value_set
                for connection_param in self.connection_params.values()
            ]
        )

    def param_exists(self, name: str) -> bool:
        return name in self.param_names

    def add_float(self, float_param: FloatParam) -> None:
        self.float_params[float_param.name] = float_param

    def add_connection(self, connection_param: ConnectionParam) -> None:
        self.connection_params[connection_param.name] = connection_param

    def print(self) -> None:
        print(f"Message {self.type}/{self.performative}")
        for float_param in self.float_params.values():
            float_param.print()
        for connection_param in self.connection_params.values():
            connection_param.print()



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Add, Cos, Divide, Multiply, Sin, Subtract

if TYPE_CHECKING:
    from aasm.parsing.state import State


def handle_math_statement(state: State, op: str, arg1: str, arg2: str) -> None:
    state.require(
        state.in_action, "Not inside any action", f"{op} can be used inside actions."
    )
    lhs = Argument(state, arg1)
    rhs = Argument(state, arg2)
    state.require(
        lhs.math_context(rhs),
        "Mismatched types in the math statement.",
        f"ARG1 {lhs.explain()}, ARG2 {rhs.explain()}",
    )

    match op:
        case "ADD":
            state.last_action.add_instruction(Add(lhs, rhs))

        case "SUBT":
            state.last_action.add_instruction(Subtract(lhs, rhs))

        case "MULT":
            state.last_action.add_instruction(Multiply(lhs, rhs))

        case "DIV":
            state.last_action.add_instruction(Divide(lhs, rhs))

        case "SIN":
            state.last_action.add_instruction(Sin(lhs, rhs))

        case "COS":
            state.last_action.add_instruction(Cos(lhs, rhs))

        case _:
            state.panic(f"Unexpected error: {op} {arg1} {arg2}")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_EBLOCK(state: State) -> None:
    state.require(state.in_action, "Cannot end blocks outside actions.")
    state.require(
        state.last_action.nested_blocks_count > 0,
        "No more blocks to close",
        "Try removing this statement.",
    )

    state.last_action.end_block()



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.parsing.op.action import op_ACTION, op_EACTION
from aasm.parsing.op.agent import op_AGENT, op_EAGENT
from aasm.parsing.op.behav import op_BEHAV, op_EBEHAV
from aasm.parsing.op.clr import op_CLR
from aasm.parsing.op.conditional import (
    handle_ordered_conditional_statement,
    handle_unordered_conditional_statement,
)
from aasm.parsing.op.decl import op_DECL
from aasm.parsing.op.defg import op_DEFG
from aasm.parsing.op.defnode import op_DEFNODE
from aasm.parsing.op.eblock import op_EBLOCK
from aasm.parsing.op.graph import op_EGRAPH, op_GRAPH
from aasm.parsing.op.len import op_LEN
from aasm.parsing.op.list_inclusion import handle_list_inclusion
from aasm.parsing.op.list_modification import handle_list_modification
from aasm.parsing.op.lr import op_LR
from aasm.parsing.op.lw import op_LW
from aasm.parsing.op.math import handle_math_statement
from aasm.parsing.op.math_exp import handle_math_exp_statement
from aasm.parsing.op.math_mod import op_MOD
from aasm.parsing.op.message import op_EMESSAGE, op_MESSAGE
from aasm.parsing.op.prm import op_agent_PRM, op_message_PRM
from aasm.parsing.op.rand import op_RAND
from aasm.parsing.op.remen import op_REMEN
from aasm.parsing.op.round import op_ROUND
from aasm.parsing.op.scale import op_SCALE
from aasm.parsing.op.send import op_SEND
from aasm.parsing.op.set import op_SET
from aasm.parsing.op.size import op_SIZE
from aasm.parsing.op.subs import op_SUBS
from aasm.parsing.state import State

if TYPE_CHECKING:
    from aasm.parsing.state import ParsedData


def parse_lines(lines: List[str], debug: bool) -> ParsedData:
    state = State(lines, debug)
    for tokens in state.tokens_from_lines():
        match tokens:
            case ["AGENT", name]:
                op_AGENT(state, name)

            case ["EAGENT"]:
                op_EAGENT(state)

            case ["MESSAGE", name, performative]:
                op_MESSAGE(state, name, performative)

            case ["EMESSAGE"]:
                op_EMESSAGE(state)

            case ["PRM", name, category]:
                op_message_PRM(state, name, category)

            case ["PRM", name, category, *args]:
                op_agent_PRM(state, name, category, args)

            case ["BEHAV", name, category, *args]:
                op_BEHAV(state, name, category, args)

            case ["EBEHAV"]:
                op_EBEHAV(state)

            case ["ACTION", name, category, *args]:
                op_ACTION(state, name, category, args)

            case ["EACTION"]:
                op_EACTION(state)

            case ["DECL", name, category, value]:
                op_DECL(state, name, category, value)

            case ["EBLOCK"]:
                op_EBLOCK(state)

            case ["IEQ" | "INEQ" | "WEQ" | "WNEQ" as op, arg1, arg2]:
                handle_unordered_conditional_statement(state, op, arg1, arg2)

            case [
                "IGT"
                | "IGTEQ"
                | "ILT"
                | "ILTEQ"
                | "WGT"
                | "WGTEQ"
                | "WLT"
                | "WLTEQ" as op,
                arg1,
                arg2,
            ]:
                handle_ordered_conditional_statement(state, op, arg1, arg2)

            case ["ADD" | "SUBT" | "MULT" | "DIV" | "SIN" | "COS" as op, arg1, arg2]:
                handle_math_statement(state, op, arg1, arg2)

            case ["LOG" | "POW" as op, arg1, arg2, arg3]:
                handle_math_exp_statement(state, op, arg1, arg2, arg3)

            case ["MOD", dst, dividend, divisor]:
                op_MOD(state, dst, dividend, divisor)

            case ["ADDE" | "REME" as op, list_, element]:
                handle_list_modification(state, op, list_, element)

            case ["LEN", result, list_]:
                op_LEN(state, result, list_)

            case ["CLR", list_]:
                op_CLR(state, list_)

            case ["IN" | "NIN" as op, list_, element]:
                handle_list_inclusion(state, op, list_, element)

            case ["SEND", rcv_list]:
                op_SEND(state, rcv_list)

            case ["SUBS", dst_list, src_list, num]:
                op_SUBS(state, dst_list, src_list, num)

            case ["SET", arg1, arg2]:
                op_SET(state, arg1, arg2)

            case ["REMEN", list_, num]:
                op_REMEN(state, list_, num)

            case ["RAND", result, cast_to, dist, *args]:
                op_RAND(state, result, cast_to, dist, args)

            case ["ROUND", num]:
                op_ROUND(state, num)

            case ["SET", arg1, arg2]:
                op_SET(state, arg1, arg2)

            case ["GRAPH", category]:
                op_GRAPH(state, category)

            case ["EGRAPH"]:
                op_EGRAPH(state)

            case ["SIZE", size]:
                op_SIZE(state, size)

            case ["SCALE", scale]:
                op_SCALE(state, scale)

            case ["DEFG", agent_name, amount, *args]:
                op_DEFG(state, agent_name, amount, args)

            case ["DEFNODE", agent_name, row]:
                op_DEFNODE(state, agent_name, row)

            case ["LR", dst, list_, idx]:
                op_LR(state, dst, list_, idx)

            case ["LW", list_, idx, value]:
                op_LW(state, list_, idx, value)

            case _:
                state.panic(f"Unknown tokens: {tokens}")

    return state.get_parsed_data()



from __future__ import annotations

from aasm.preprocessor.preprocessor_item import PreprocessorItem


class Constant(PreprocessorItem):
    def __init__(self, signature: str):
        super().__init__(signature)
        self.name = ""

    def __repr__(self):
        return f"Constant({self.name})"

    def add_definition(self, name: str, value: str):
        self.name = name
        self.value = value

    def expand(self, line: str) -> str:
        tokens = [token.strip() for token in line.strip().replace(",", " ").split()]
        tokens = [self.value if token == self.name else token for token in tokens]
        if len(tokens) < 2:
            return line
        new_line = tokens[0] + " "
        new_line = new_line + " ".join(tok + "," for tok in tokens[1:])
        new_line = new_line.lstrip(",")
        return new_line



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.action import ModifySelfAction, SendMessageAction
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_ACTION(state: State, name: str, category: str, args: List[str]) -> None:
    state.require(
        state.in_behaviour,
        "Actions must be definied inside behaviours.",
        "Try defining new behaviours using BEHAV.",
    )
    state.require(
        not state.in_action,
        "Cannot define an action in another action.",
        "Try finishing current action using EACTION.",
    )
    state.require(
        not state.last_behaviour.action_exists(name),
        f"Action {name} already exists in current behaviour.",
    )
    state.require(
        is_valid_name(name),
        f"{name} is not a correct name.",
        f"Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.",
    )

    match category, args:
        case "modify_self", []:
            state.last_behaviour.add_action(ModifySelfAction(name))

        case "send_msg", [msg_type, msg_performative]:
            state.require(
                state.message_exists(msg_type, msg_performative),
                f"Message {msg_type}/{msg_performative} does not exist.",
                "Try defining new messages using MESSAGE.",
            )

            state.last_behaviour.add_action(
                SendMessageAction(
                    name, state.get_message_instance(msg_type, msg_performative)
                )
            )

        case _:
            state.panic(f"Incorrect operation: ACTION {category} {args}")

    state.in_action = True


def op_EACTION(state: State) -> None:
    state.require(
        state.in_action,
        "Not inside any action.",
        "Try defining new actions using ACTION.",
    )
    state.require(
        state.last_action.nested_blocks_count == 0,
        "There are unclosed blocks.",
        "Try closing them using EBLOCK.",
    )

    state.in_action = False



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import (
    IfEqual,
    IfGreaterThan,
    IfGreaterThanOrEqual,
    IfLessThan,
    IfLessThanOrEqual,
    IfNotEqual,
    WhileEqual,
    WhileGreaterThan,
    WhileGreaterThanOrEqual,
    WhileLessThan,
    WhileLessThanOrEqual,
    WhileNotEqual,
)

if TYPE_CHECKING:
    from aasm.parsing.state import State


def handle_unordered_conditional_statement(
    state: State, op: str, arg1: str, arg2: str
) -> None:
    state.require(
        state.in_action, "Not inside any action.", f"{op} can be used inside actions."
    )
    lhs = Argument(state, arg1)
    rhs = Argument(state, arg2)
    state.require(
        lhs.unordered_comparaison_context(rhs),
        "Mismatched types in the unordered comparaison context.",
        f"ARG1 {lhs.explain()}, ARG2 {rhs.explain()}",
    )

    match op:
        case "IEQ":
            state.last_action.add_instruction(IfEqual(lhs, rhs))

        case "INEQ":
            state.last_action.add_instruction(IfNotEqual(lhs, rhs))

        case "WEQ":
            state.last_action.add_instruction(WhileEqual(lhs, rhs))

        case "WNEQ":
            state.last_action.add_instruction(WhileNotEqual(lhs, rhs))

        case _:
            state.panic(f"Unexpected error: {op} {arg1} {arg2}")

    state.last_action.start_block()


def handle_ordered_conditional_statement(
    state: State, op: str, arg1: str, arg2: str
) -> None:
    state.require(
        state.in_action, "Not inside any action.", f"{op} can be used inside actions."
    )
    lhs = Argument(state, arg1)
    rhs = Argument(state, arg2)
    state.require(
        lhs.ordered_comparaison_context(rhs),
        "Mismatched types in the ordered comparaison context.",
        f"ARG1 {lhs.explain()}, ARG2 {rhs.explain()}",
    )

    match op:
        case "IGT":
            state.last_action.add_instruction(IfGreaterThan(lhs, rhs))

        case "IGTEQ":
            state.last_action.add_instruction(IfGreaterThanOrEqual(lhs, rhs))

        case "ILT":
            state.last_action.add_instruction(IfLessThan(lhs, rhs))

        case "ILTEQ":
            state.last_action.add_instruction(IfLessThanOrEqual(lhs, rhs))

        case "WGT":
            state.last_action.add_instruction(WhileGreaterThan(lhs, rhs))

        case "WGTEQ":
            state.last_action.add_instruction(WhileGreaterThanOrEqual(lhs, rhs))

        case "WLT":
            state.last_action.add_instruction(WhileLessThan(lhs, rhs))

        case "WLTEQ":
            state.last_action.add_instruction(WhileLessThanOrEqual(lhs, rhs))

        case _:
            state.panic(f"Unexpected error: {op} {arg1} {arg2}")

    state.last_action.start_block()



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.message import Message
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_MESSAGE(state: State, msg_type: str, msg_performative: str) -> None:
    state.require(
        not state.in_message,
        "Already inside a message.",
        "First end current message using EMESSAGE.",
    )
    state.require(
        not state.in_agent,
        "Cannot define messages inside agents.",
        "First end current agent using EAGENT.",
    )
    state.require(
        not state.in_graph,
        "Cannot define messages inside graphs.",
        "First end current graph using EGRAPH.",
    )
    state.require(
        not state.message_exists(msg_type, msg_performative),
        f"Message {msg_type}/{msg_performative} already exists in the current environment.",
    )
    state.require(
        is_valid_name(msg_type),
        f"{msg_type} is not a correct name.",
        f"Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.",
    )
    state.require(
        is_valid_name(msg_performative),
        f"{msg_performative} is not a correct name.",
        f"Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.",
    )

    state.in_message = True
    state.add_message(Message(msg_type, msg_performative))


def op_EMESSAGE(state: State) -> None:
    state.require(
        state.in_message,
        "Not inside any message.",
        "Try defining new messages using MESSAGE.",
    )

    state.in_message = False



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.action import SendMessageAction
from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Send

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_SEND(state: State, arg1: str) -> None:
    state.require(
        state.in_action,
        "Not inside any action.",
        "SEND can be used inside send_msg actions.",
    )
    state.require(
        isinstance(state.last_action, SendMessageAction),
        "Not inside send_msg action.",
        "SEND can be used inside send_msg actions.",
    )
    receivers = Argument(state, arg1)
    state.require(
        receivers.send_context(),
        "Mismatched type in the send context.",
        f"ARG1 {receivers.explain()}",
    )

    state.last_action.add_instruction(Send(receivers))



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import Logarithm, Power

if TYPE_CHECKING:
    from aasm.parsing.state import State


def handle_math_exp_statement(
    state: State, op: str, arg1: str, arg2: str, arg3: str
) -> None:
    state.require(
        state.in_action, "Not inside any action", f"{op} can be used inside actions."
    )
    dst = Argument(state, arg1)
    base = Argument(state, arg2)
    arg = Argument(state, arg3)
    state.require(
        dst.math_exponentiation_context(base, arg),
        "Mismatched types in the math exponentiation statement.",
        f"ARG1 {dst.explain()}, ARG2 {base.explain()}, ARG3 {arg.explain()}",
    )

    match op:
        case "LOG":
            state.last_action.add_instruction(Logarithm(dst, base, arg))

        case "POW":
            state.last_action.add_instruction(Power(dst, base, arg))

        case _:
            state.panic(f"Unexpected error: {op} {arg1} {arg2} {arg3}")



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.graph import AdjRow, MatrixAgent, MatrixGraph

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_DEFNODE(state: State, agent_name: str, agent_row: str) -> None:
    state.require(state.agent_exists(agent_name), f"Agent {agent_name} is not defined.")
    state.require(
        state.in_graph,
        "Cannot define agent node outside of graph scope.",
        "Try defining new graph using GRAPH.",
    )
    state.require(
        isinstance(state.last_graph, MatrixGraph),
        "DEFNODE can be used only in matrix graphs.",
        "Define matrix graphs with GRAPH matrix.",
    )
    state.require(
        agent_row.startswith("R"), "Agent Row must start with R", "Example: R101"
    )
    row = agent_row[1:]
    row = [int(x) for x in row]
    adj_row = AdjRow(row)

    state.last_graph.add_agent(MatrixAgent(agent_name, adj_row))



from __future__ import annotations

from typing import List, Tuple

from aasm.preprocessor.constant import Constant
from aasm.preprocessor.macro import Macro
from aasm.utils.exception import PanicException


class Preprocessor:
    def __init__(self, lines: List[str]):
        self.lines = lines
        self.ignore = []
        self.processed_lines = lines.copy()
        self.macros = []
        self.constants = []
        self.line_expansions = []
        self.ignore_offsets = []
        self.macro_offsets = []
        self.item_names = []
        self.post_ignore = {}
        self.post_macros = {}

    def get_original_line_number(self, line_idx: int) -> Tuple[int, str]:
        for exp in self.line_expansions:
            if line_idx == exp[0]:
                return (exp[1].declare_line, f"MAKRO call: {exp[1].name}")
        post_macro_idx = self.post_macros[line_idx]
        return (self.post_ignore[post_macro_idx], "")

    def get_makro_name(self, line_idx: int) -> str:
        line = self.lines[line_idx - 1]
        tokens = [token.strip() for token in line.strip().replace(",", " ").split()]
        if len(tokens) > 0:
            m_name = tokens[0]
            if m_name in [macro.name for macro in self.macros]:
                return m_name
        return ""

    def run(self):
        self.parse_items()
        self.clean_ignored()
        self.expand_constants()
        self.expand_macros()
        return self.processed_lines

    def clean_ignored(self):
        for idx in sorted(self.ignore, reverse=True):
            for offset_idx in range(len(self.ignore_offsets)):
                if offset_idx >= idx:
                    self.ignore_offsets[offset_idx] -= 1
            del self.processed_lines[idx]
        self.ignore_offsets = list(reversed(self.ignore_offsets))

    def expand_constants(self):
        line_idx = 0
        for line in self.processed_lines:
            expanded = line
            for const in self.constants:
                expanded = const.expand(expanded)
            self.processed_lines[line_idx] = expanded
            line_idx += 1

    def expand_macros(self):
        line_idx = 0
        to_expand = []
        names = [macro.name for macro in self.macros]
        for line in self.processed_lines:
            tokens = [token.strip() for token in line.strip().replace(",", " ").split()]
            if len(tokens) != 0 and tokens[0] in names:
                to_expand.append((line_idx, tokens[0], tokens[1:]))
            line_idx += 1

        offset = 0
        makro_lines = []
        for makro in to_expand:
            line_idx, name, args = makro
            line_idx += offset
            macro_item = self.macros[names.index(name)]
            if len(makro[2]) != len(macro_item.argument_regexes):
                raise PanicException(
                    f"Error in line: {line_idx}",
                    "Wrong number of arguments",
                    f"Expected {len(macro_item.argument_regexes)} arguments, got {len(args)}",
                )
            self.processed_lines[line_idx:line_idx] = macro_item.expand(args)
            for offset_idx in range(macro_item.expand_len - 1):
                self.macro_offsets.append(line_idx + offset_idx + 2)
            del self.processed_lines[line_idx + macro_item.expand_len]
            for line_offset in range(macro_item.expand_len - 1):
                makro_lines.append(line_idx + line_offset + 2)
            self.line_expansions.append((line_idx + 1, macro_item))
            offset += macro_item.expand_len - 1

        line_idx = 0
        post_idx = 0
        for line in self.processed_lines:
            post_idx += 1
            if post_idx not in makro_lines:
                line_idx += 1
            self.post_macros[post_idx] = line_idx

    def parse_items(self):
        line_idx = 0
        currentItem = None
        post_ignore_idx = 1
        for line in self.lines:
            self.ignore_offsets.append(0)
            self.post_ignore[post_ignore_idx] = line_idx + 1
            line_idx += 1
            tmp = line.strip()
            # enter preprocessor directive
            if len(tmp) == 0:
                self.ignore.append(line_idx - 1)
                continue
            elif tmp[0] == "%":
                self.ignore.append(line_idx - 1)
                signature = tmp.lstrip("%")
                tokens = [
                    token.strip() for token in signature.replace(",", " ").split()
                ]
                if tokens:
                    tokens[0] = tokens[0].upper()
                match tokens:
                    case ["MAKRO", *makro_def]:
                        if currentItem is None:
                            currentItem = Macro(signature)
                            currentItem.add_definition(makro_def, line_idx)
                            if currentItem.name in self.item_names:
                                raise PanicException(
                                    f"Error in line: {line_idx}",
                                    "Duplicate preprocessor item name",
                                    currentItem.name,
                                )
                            self.item_names.append(currentItem.name)
                        else:
                            raise PanicException(
                                f"Error in line: {line_idx}",
                                "Nested preprocessor directives!",
                                "Make sure that you don't use preprocessor directives inside each other.",
                            )
                    case ["EMAKRO"]:
                        if isinstance(currentItem, Macro):
                            self.macros.append(currentItem)
                            currentItem = None
                        else:
                            raise PanicException(
                                f"Error in line: {line_idx}",
                                "Closing a makro without opening one!",
                                "Add a matching %MAKRO directive",
                            )
                        pass
                    case ["CONST", const_name, const_value]:
                        if currentItem is None:
                            currentItem = Constant(signature)
                            currentItem.add_definition(const_name, const_value)
                            self.constants.append(currentItem)
                            if currentItem.name in self.item_names:
                                raise PanicException(
                                    f"Error in line: {line_idx}",
                                    "Duplicate preprocessor item name",
                                    currentItem.name,
                                )
                            self.item_names.append(currentItem.name)
                            currentItem = None
                        else:
                            raise PanicException(
                                f"Error in line: {line_idx}",
                                "Nested preprocessor directive!",
                                "Make sure that you don't use preprocessor directives inside each other",
                            )
                    case _:
                        raise PanicException(
                            f"Error in line: {line_idx}",
                            "Unknown preprocessor directive",
                            "Remove '%' from beggining of the line",
                        )
            elif tmp[0] == "#":
                self.ignore.append(line_idx - 1)
            elif currentItem is not None:
                self.ignore.append(line_idx - 1)
                if isinstance(currentItem, Macro):
                    currentItem.add_line(line)
            else:
                # The case where the line is not ignored
                post_ignore_idx += 1



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.declaration import ConnectionDeclaration, FloatDeclaration
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_DECL(state: State, name: str, category: str, value: str) -> None:
    state.require(state.in_action, "Cannot declare variables outside actions.")
    state.require(
        is_valid_name(name),
        f"{name} is not a correct name.",
        f"Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.",
    )
    state.require(
        not state.last_agent.param_exists(name),
        f"{name} is already defined in current agent.",
    )
    state.require(
        not state.last_action.is_declaration_in_scope(name),
        f"{name} is already declared in current action scope.",
    )
    name_arg = Argument(state, name)
    value_arg = Argument(state, value)
    state.require(
        name_arg.declaration_context(value_arg),
        "Mismatched types in the declaration context.",
        f"NAME {name_arg.explain()}, VALUE {value_arg.explain()}",
    )

    match category:
        case "float":
            state.last_action.add_float_declaration(
                FloatDeclaration(name_arg, value_arg)
            )

        case "conn":
            state.last_action.add_connection_declaration(
                ConnectionDeclaration(name_arg, value_arg)
            )

        case _:
            state.panic(f"Incorrect declaration: DECL {name} {category} {value}")
